/*
 * What to do:
 *
 *  gpg-decrypt keyfile
 *  split by lines into keys (64/65)
 *  mode detect on numkeys: v2/v3
 *
 *  foreach keystr 
 *    keybits[n] = truncate(HASH(keystr))
 *    keybits[n][0] =^ modemagic (v3: 0xF4, v2: 0x55)
 *
 *  HASH is 
 *    SHA-256() for AES-128
 *    SHA-384() for AES-192
 *    SHA-512() for AES-256
 *
 *  HASH output is truncated to keysize of cipher
 */

#include <stdio.h>
#include <unistd.h>
#include <getopt.h>
#include <sys/mman.h>

#include <gcrypt.h>

int gcrypt_init(void)
{
	/* Version check should be the very first call because it
	   makes sure that important subsystems are intialized. */
	if (!gcry_check_version(GCRYPT_VERSION)) {
		fputs ("libgcrypt version mismatch\n", stderr);
		return 2;
	}

	/* ... If required, other initialization goes here.  Note that the
	   process might still be running with increased privileges and that
	   the secure memory has not been intialized.  */

	/* Allocate a pool of 16k secure memory.  This make the secure memory
	   available and also drops privileges where needed.  */
	gcry_control(GCRYCTL_INIT_SECMEM, 16384, 0);

	/* It is now okay to let Libgcrypt complain when there was/is
	   a problem with the secure memory. */
	gcry_control(GCRYCTL_RESUME_SECMEM_WARN);

	/* ... If required, other initialization goes here.  */

	/* Tell Libgcrypt that initialization has completed. */
	gcry_control(GCRYCTL_INITIALIZATION_FINISHED, 0);

	return 0;
}

/* largest: SHA-512 */
#define DIGEST_MAXSIZE 64
#define KEY_MAXSIZE DIGEST_MAXSIZE

#define IVKEY_MAXSIZE 16

struct multikey {
	int numkeys;
	unsigned char key[65][DIGEST_MAXSIZE];
	unsigned char ivkey[IVKEY_MAXSIZE];
};

static int multikey_version(struct multikey *key)
{
	/* Version 2 uses 64 keys */
	if (key->numkeys == 64)
		return 2;

	/* Version 3 uses 65 keys */
	if (key->numkeys == 65)
		return 3;

	return -1;
}

typedef int (*digestfunc_t)(const char *in, size_t inlen, unsigned char *out);

int multikey_read(struct multikey *multikey, const char *filename, digestfunc_t digest)
{
	FILE *f;
	char line[128];
	unsigned int numkey = 0;
	unsigned char tweak;

	f = fopen(filename, "r");
	if (!f) {
		perror(filename);
		return -1;
	}

	numkey = 0;
	while (numkey < 65) {
		size_t len;
	
		/* Read line of key material */
		if (!fgets(line, sizeof(line), f)) {
			if (ferror(f))
				goto bad_input;
			
			if (feof(f)) {
				/* v2 key has 64 lines */
				if (numkey == 63)
					break;
				/* v3 key has 65 lines */
				if (numkey == 64)
					break;
			}

			goto bad_input;
		}

		len = strlen(line);

		/* Too short? */
		if (len < 2)
			goto bad_input;
		
		/* Truncated line? */
		if (line[len-1] != '\n')
			goto bad_input;

		/* Strip newline */
		line[--len] = 0;
	
		/* Warn? */
		if (len < 60)
			fprintf(stderr, "Warning: unusually short line (len=%zu).\n", len);
		if (len > 60)
			fprintf(stderr, "Warning: unusually long lines (len=%zu).\n", len);

		/* Digest it */
		if (digest(line, len, multikey->key[numkey]))
			goto fail;

		/* Wipe line */
		memset(line, 0, sizeof(line));

		++numkey;
	}

	/* Now: We expect to have either 64 or 65 keys */
	if (numkey < 64 || numkey > 65)
		goto bad_input;

	multikey->numkeys = numkey;

	/* Tweak:
	 */

	if (numkey == 64)
		tweak = 0x55;
	else
		tweak = 0xF4;

	while (numkey--)
		multikey->key[numkey][0] ^= tweak;

	return 0;

fail:
bad_input:
	memset(line, 0, sizeof(line));
	memset(multikey, 0, sizeof(*multikey));

	return -1;
}

int __digest(int algo, const char *in, size_t inlen, unsigned char *out, size_t outlen)
{
	gcry_md_hd_t hash;
	gcry_error_t gerr;

	gerr = gcry_md_open(&hash, algo, GCRY_MD_FLAG_SECURE);
	if (gerr)
		return 1;

	gcry_md_write(hash, in, inlen);

	memcpy(out, gcry_md_read(hash, 0), outlen);

	gcry_md_close(hash);

	return 0;
}

/*
 * Note: The "outsizes" below are smaller than the output
 * of the respective hash functions.
 *
 * This is because loop-AES losetup truncates their output
 * to form 128-256 bit encryption keys.
 *
 */

int digest_sha256(const char *in, size_t inlen, unsigned char *out)
{
	return __digest(GCRY_MD_SHA256, in, inlen, out, 16);
}

int digest_sha384(const char *in, size_t inlen, unsigned char *out)
{
	return __digest(GCRY_MD_SHA384, in, inlen, out, 24);
}

int digest_sha512(const char *in, size_t inlen, unsigned char *out)
{
	return __digest(GCRY_MD_SHA512, in, inlen, out, 32);
}

void usage(const char *argv0)
{
	fprintf(stderr,
		"usage: %s [options] <keyfile>\n" \
		"options\n" \
		"  -f <format>      (dm-crypt, raw)\n" \
		"  -d <digestsize>  In bits (256, 384, 512)\n",
		argv0);
}

int main(int argc, char **argv)
{
	const char *keyfile;
	int keysize;
	struct multikey multikey;
	enum { KEYFORMAT_RAW, KEYFORMAT_DMCRYPT } keyformat;
	int i, j;
	digestfunc_t digestfunc;
	int digestsize;
	int opt;

	while ((opt = getopt(argc, argv, "d:f:h")) != -1) {
		switch (opt) {
		case 'd':
			digestsize = atoi(optarg);
			break;
		case 'f':
			if (strcmp(optarg, "dm-crypt") == 0)
				keyformat = KEYFORMAT_DMCRYPT;
			else if (strcmp(optarg, "raw") == 0)
				keyformat = KEYFORMAT_RAW;
			else {
				fprintf(stderr, "Unsupported keyformat '%s'\n", optarg);
				exit(1);
			}
			break;
		case 'h':
			usage(argv[0]);
			exit(0);
		}
	}

	if (argc - optind == 0)
		keyfile = "/dev/stdin";
	else if (argc - optind == 1)
		keyfile = argv[optind];
	else {
		usage(argv[0]);
		exit(1);
	}

	switch (digestsize) {
	case 256:
		digestfunc = digest_sha256;
		break;
	case 384:
		digestfunc = digest_sha384;
		break;
	case 512:
		digestfunc = digest_sha512;
		break;
	default:
		fprintf(stderr, "Error: Unsupported digest size %d\n", digestsize);
		exit(1);
	}

	/*
	 * Actual keysize; truncated
	 */
	keysize = (digestsize/8)/2;

	if (gcrypt_init())
		exit(1);
	
	memset(&multikey, 0, sizeof(multikey));

	/*
	 * Prevent 
	 *
	* XXX mlock() */

	mlock(&multikey, sizeof(multikey));

	if (multikey_read(&multikey, keyfile, digestfunc)) {
		memset(&multikey, 0, sizeof(multikey));
		exit(1);
	}

	switch (keyformat) {
	case KEYFORMAT_RAW:
		/*
		 * Raw keys, can be used with loop-AES and
		 * ioctl(LOOP_SET_MULTIKEY)
		 * ioctl(LOOP_SET_MULTIKEY_V3)
		 *
		 * The keys are each zero-padded to 32 byte.
		 */
		for (i=0; i < multikey.numkeys; ++i) {
			write(1, multikey.key[i], keysize);
			if (keysize < 32)
				for (j=0; j < 32-keysize; j++)
					write(1, "\x00", 1);
		}
		break;

	case KEYFORMAT_DMCRYPT:
		/*
		 * Hex string, can be used with dmsetup for
		 * dm-crypt mode ruusu2/3-sector-multi:64.
		 */

		if (multikey_version(&multikey) == 3) {
			/* 
			 * For v3, which has an extra IV key, we
			 * output it as series of
			 *   <key><ivkey>
			 *   <key><ivkey>
			 *
			 * The IV key is duplicated.
			 */
	
			for (i=0; i < 64; i++) {

				/* XXX Gn. Should not pass through stdio 
				 * like that. Rather:
				 *
				 *  Allocate mlocked() buffer.
				 *  Build hexstring in buffer.
				 *  Pass hexstring to write(keyoutfd)
				 */
			
				/*
				 * So lets build string I can use with
				 * dm-crypt ruusu3.
				 */

				for (j=0; j < keysize; j++)
					fprintf(stdout, "%02x", multikey.key[i][j]);

				for (j=0; j < 16; j++)
					fprintf(stdout, "%02x", multikey.key[64][j]);
			}

		} else {
			for (i=0; i < multikey.numkeys; ++i)
				for (j=0; j < keysize; j++)
					fprintf(stdout, "%02x", multikey.key[i][j]);
		}
		break;
	}

	memset(&multikey, 0, sizeof(multikey));

	/* Wipe secure memory */
	gcry_control(GCRYCTL_TERM_SECMEM);

	return 0;
}
